﻿-- SpellReport addon @ 2018 MiChaeL
local addonName = ...

local addon = CreateFrame("Frame", addonName )

-- caching some variables and functions
local type = type
local print = print
local band = bit.band
local fmt = string.format
local gsub = string.gsub
local IsInRaid = IsInRaid
local IsInGroup = IsInGroup
local GetSpellLink = GetSpellLink
local GetSpellInfo = GetSpellInfo
local SendChatMessage = SendChatMessage
local GetNumGroupMembers = GetNumGroupMembers
local CombatLogGetCurrentEventInfo = CombatLogGetCurrentEventInfo
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local PLAYER_NAME  = UnitName('player')
local PLAYER_GUID  = UnitGUID('player')

-- built-in categories
local CATEGORIES_BUILTIN = {
	"TAUNT", "INTERRUPT", "DISPEL", "STEAL", "PORTAL", "FOOD", "HEALTH",
	PORTAL    = "SPELL_CAST_START",
	FOOD      = "SPELL_CAST_START",
	HEALTH    = "SPELL_CAST_START",
	TAUNT     = "SPELL_CAST_SUCCESS",
	INTERRUPT = "SPELL_INTERRUPT",
	DISPEL    = "SPELL_DISPEL",
	STEAL     = "SPELL_STOLEN",
}

-- CATEGORIES_BUILTIN + custom categories
local CATEGORIES 

local CATEGORY_MASKS = {
	PORTAL    = "casting {spell}",
	FOOD      = "casting {spell}",
	HEALTH    = "casting {spell}",
	TAUNT     = "taunted {dst} using {spell}", 
	INTERRUPT = "interrupted {dst}'s {extraspell}",
	DISPEL    = "dispelled {dst}'s {extraspell}", 
	STEAL     = "stolen {dst}'s {extraspell}",
}

local SPELLS = {
	TAUNT = {
		355,     -- Warrior, Taunt
		51399,   -- DK, Death Grip for Blood
		56222,   -- DK, Dark Command
		62124,   -- Paladin, Hand of Reckoning
		204079,  -- Paladin, Final Stand	
		6795,    -- Druid, Growl
		115546,  -- Monk, Provoke
		185245,  -- DH, Torment
		20736,   -- Hunter, Distracting Shot
	},
	PORTAL = {
		10059,   -- STORMWIND
		11416,   -- IRONFORGE
		11417,   -- ORGRIMMAR
		11418,   -- UNDERCITY
		11419,   -- DARNASSUS
		11420,   -- THUNDER BLUFF
		49360,   -- THERAMORE
		49361,   -- STONARD
		120146,  -- ANCIENT DALARAN
		32266,   -- EXODAR
		32267,   -- SILVERMOON
		33691,   -- SHATTRATH
		35717,   -- SHATTRATH
		53142,   -- DALARAN NORTHREND
		88345,   -- TOL BARAD
		88346,   -- TOL BARAD			
		132620,  -- VALE OF ETERNAL BLOSSOMS
		132626,  -- VALE OF ETERNAL BLOSSOMS
		176244,  -- WARSPEAR
		176246,  -- STORMSHIELD
		224871,  -- DALARAN BROKEN ISLES
		281400,  -- BORALUS
		281402,  -- DAZAR'ALOR
	},
	FOOD = {
		190336,  -- Refreshment table, Mage
	},
	HEALTH = {
		29893,   -- Soulwell, Warlock
	},
}	

-- format message
local Format
do
	local function fmtName(name, guid)
		if guid then
			local _,class = GetPlayerInfoByGUID(guid)
			if class then
				return fmt("|c%s%s|r", RAID_CLASS_COLORS[class].colorStr, name)
			end
		end
		return name
	end
	local function fmtSpell(spellID, dest)
		return dest~='say' and GetSpellLink(spellID) or GetSpellInfo(spellID)
	end
	function Format(dest, category, srcGUID, srcName, dstGUID, dstName, spellID, extraSpellID)
		if dest ~= 'me' then
			srcGUID, dstGUID = nil, nil -- chat channels don't allow escape characters, so do not format names with class color
		end
		local msg = CATEGORY_MASKS[category] or addon.db.messages[spellID] or addon.db.messages[category] or (dstName and "casted {spell} over {dst}") or "casted {spell}" 
		if spellID then
			msg = gsub( msg, "{spell}", fmtSpell(spellID,dest) )
		end
		if extraSpellID then 
			msg = gsub( msg, "{extraspell}", fmtSpell(extraSpellID,dest) ) 
		end
		if dstName then 
			msg = gsub( msg, "{dst}", fmtName(dstName,dstGUID) )
		end
		if srcName~=PLAYER_NAME then 
			msg = fmt( "%s %s", fmtName(srcName,srcGUID), msg )
		elseif dest=='me' then
			msg = fmt( "[%s]: %s", fmtName(srcName,srcGUID), msg )
		end
		return msg
	end
end

-- report message to chat channel
local function Report(msg, dest)
	if dest=='me' then
		print(msg)
	elseif dest=='say' then
		SendChatMessage(msg, 'SAY' )
	else
		SendChatMessage(msg, (IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and "INSTANCE_CHAT") or (IsInRaid() and "RAID" or "PARTY"))
	end
end

-- spells tracking and reporting 
local UpdateSpells, CombatLogEvent
do
	--populate events table with enabled categories and spells
	local events = {}
	function UpdateSpells()
		CATEGORIES = CopyTable(CATEGORIES_BUILTIN)
		for spell,cat in pairs(addon.db.spells) do
			if not CATEGORIES[cat] then
				CATEGORIES[cat] = 'SPELL_CAST_SUCCESS'
				table.insert( CATEGORIES, cat)
				SPELLS[cat] = {}
			end
			table.insert( SPELLS[cat], spell )
		end
		local disabled = addon.db.disabled
		wipe(events)
		for _,cat in ipairs(CATEGORIES) do
			if not disabled[cat] then
				local event = CATEGORIES[cat]
				if SPELLS[cat] then
					local spells = events[event] or {}
					for _,spellID in ipairs(SPELLS[cat]) do
						spells[spellID] = cat
					end
					events[event] = spells
				else
					events[event] = cat
				end
			end
		end
	end
	-- combatlog event 
	local MINE_FLAGS  = COMBATLOG_OBJECT_AFFILIATION_MINE
	local GROUP_FLAGS = COMBATLOG_OBJECT_AFFILIATION_RAID + COMBATLOG_OBJECT_AFFILIATION_PARTY 
	function CombatLogEvent()
		local _, event, _, srcGUID, srcName, srcFlags, _, dstGUID, dstName, _, dstFlags, spellID, _ , _, extraSpellID = CombatLogGetCurrentEventInfo()
		local cat = events[event]
		if cat then
			if type(cat)=="table" then
				cat = cat[spellID]
				if not cat then return end
			end
			local src = ((srcName==PLAYER_NAME or band(srcFlags, MINE_FLAGS)~=0) and "me") or (band(srcFlags, GROUP_FLAGS)~=0 and "others")
			if src then -- track only me or people in my group/raid
				local dbsrc = addon.db.source[cat] or "all"
				if dbsrc=="all" or src==dbsrc then
					local dst = addon.db.report[cat] or 'me'
					if GetNumGroupMembers()==0 then
						dst = dst~='others' and 'me'
						if not dst then return end
					end
					local msg = Format( dst, cat, srcGUID, srcName, dstGUID, dstName, spellID, extraSpellID )
					Report( msg, dst )
				end
			end
		end	
	end
end

-- config mode to capture custom spellIDs
local function CombatLogEventConfigMode()
	local _, event, _, srcGUID, srcName, srcFlags, _, dstGUID, dstName, _, dstFlags, spellID, _ , _, extraSpellID = CombatLogGetCurrentEventInfo()
	if srcGUID == PLAYER_GUID  and (event == 'SPELL_CAST_FAILED' or event == "SPELL_CAST_SUCCESS") then
		local cat = addon.db.spells[spellID]
		if cat  then
			addon.db.spells[spellID] = nil
			print( string.format( "|cff9999ffSpellReport:|r Spell %s removed from the category %s.|r", GetSpellLink(spellID), cat) )
		else
			addon.db.spells[spellID] = addon.configCategory
			print( string.format( "|cff9999ffSpellReport:|r Spell %s added to the category %s.|r", GetSpellLink(spellID), addon.configCategory ) )
		end
		UpdateSpells()
	end
end

-- load addon
addon:RegisterEvent("ADDON_LOADED")
addon:SetScript("OnEvent", function(self, event, name)
	if name == addonName then
		self:UnregisterEvent("ADDON_LOADED")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:SetScript("OnEvent", CombatLogEvent)
		SpellReportDB = SpellReportDB or { report = {}, source = {}, disabled = {}, spells = {}, messages = {} }
		self.db = SpellReportDB
		self.db.spells = self.db.spells or {}
		self.db.messages = self.db.messages or {}
		UpdateSpells()
	end	
end)

-- command line configuration
do
	local DEFAULT = { report = "me", source = "all" }
	local SVALUES  = { me = 'others', others = 'all', all = 'me' }
	local DVALUES  = { me = 'others', others = 'all', all = 'say', say = 'me' }
	local function cmdHandler(msg)
		local db = addon.db
		local cmd, par1, par2 = strsplit(" ", msg, 3)
		local cat = strupper(cmd)
		if CATEGORIES[cat] then
			if par1=='source' then
				db[par1][cat] = SVALUES[par2] and par2 or SVALUES[db[par1][cat] or DEFAULT[par1]]
			elseif par1=='report' then
				db[par1][cat] = DVALUES[par2] and par2 or DVALUES[db[par1][cat] or DEFAULT[par1]]
			else
				db.disabled[cat] = (not db.disabled[cat]) or nil
			end
		elseif DEFAULT[cmd] and VALUES[par1] then
			for _,cat in ipairs(CATEGORIES) do
				db[cmd][cat] = par1
			end
		elseif cat == 'CONFIG' then
			local category = strupper( par1 or 'custom' )
			if not CATEGORIES_BUILTIN[category] then
				addon.configCategory = category
				addon.configMode = not addon.configMode
				if addon.configMode then
					addon:SetScript('OnEvent', CombatLogEventConfigMode)
					print( string.format( "|cff9999ffSpellReport:|r Config Mode Enabled for category: [%s]. Spells casted will be registered to the specified category.", addon.configCategory) )
				else
					addon:SetScript('OnEvent', CombatLogEvent)
					print("|cff9999ffSpellReport:|r Config Mode Disabled.")
				end
			else
				print("|cff9999ffSpellReport:|r You must specify a user defined category name, built-in categories are not configurable.")
			end
			return
		elseif cat == 'MESSAGE' then
			par1 = strupper( par1 or 'custom' )
			if CATEGORIES[par1] and not CATEGORIES_BUILTIN[par1] then
				addon.db.messages[par1] = par2
			else
				local spellID = tonumber(par1)
				if spellID and addon.db.spells[spellID] then
					addon.db.messages[spellID] = par2
				end
			end
			return
		elseif cat =='SPELLS' then
			print("|cff9999ffSpellReport:|r User Defined Categories & Spells:")
			for spell,cat in pairs(addon.db.spells) do
				print( string.format( "|cff9999ffSpellReport:|r %s: %s [%d] %s", cat, (GetSpellInfo(spell)) , spell, addon.db.messages[spell] or addon.db.messages[cat] or '' ) )
			end
			return
		end
		print("|cff9999ffSpellReport:|r SpellReport by MiChaeL - Commands:")
		print("|cff9999ffSpellReport:|r     /sr source me||others||all  - Which source players to monitorize.")
		print("|cff9999ffSpellReport:|r     /sr report me||others||all||say  - Report destination channel.")
		print("|cff9999ffSpellReport:|r     /sr <category>  - Toggles category reporting on or off.")
		print("|cff9999ffSpellReport:|r     /sr <category> source me||others||all  - Which source players to monitorize.")
		print("|cff9999ffSpellReport:|r     /sr <category> report me||others||all||say  - Report destination channel.")
		print("|cff9999ffSpellReport:|r     /sr config <category> - Toggle config mode for custom category.")
		print("|cff9999ffSpellReport:|r     /sr message <category|spellID> <text> - Text to display for spell or category.")		
		print("|cff9999ffSpellReport:|r     /sr spells - Display custom spells.")
		print("|cff9999ffSpellReport:|r Categories:")
		for _,cat in ipairs(CATEGORIES) do
			print( fmt("|cff9999ffSpellReport:|r     %s: |c%s|r  source: |cFF0070dd%s|r  report: |cFF3fc6ea%s|r", strlower(cat), db.disabled[cat] and "FFE02020off" or "FF20E020on", db.source[cat] or "all", db.report[cat] or "me" ) )
		end
		UpdateSpells()
	end
	SLASH_SPELLREPORT1 = '/spellreport'
	SLASH_SPELLREPORT2 = '/sr'
	SlashCmdList.SPELLREPORT = cmdHandler
end
